@@ -15,7 +15,7 @@
 // Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 // 
 // As a special exception to the GNU General Public License, permission is 
-// granted for additional uses of the text contained in its release 
+// granted for additional uses of the text contained in its release
 // of ccaudio.
 //
 // The exception is that, if you link the ccaudio library with other
@@ -49,6 +49,8 @@
 namespace ost {
 #endif
 
+#define	AUDIO_SIGNED_LINEAR_RAW	1
+
 /**
  * Generic audio class to hold master data types and various useful
  * class encapsulated friend functions as per GNU Common C++ 2 coding
@@ -57,7 +59,7 @@ namespace ost {
  * @author David Sugar <dyfet@ostel.com>
  * @short Master audio class.
  */
-class Audio
+class CCXX_CLASS_EXPORT Audio
 {
 public:
 	enum	Rate
@@ -83,6 +85,10 @@ public:
 		alawAudio,
 		okiADPCM,
 		voxADPCM,
+
+		// Please keep the PCM types at the end of the list -
+		// see the "is this PCM or not?" code in
+		// AudioFile::close for why.
 		cdaStereo,
 		cdaMono,
 		pcm8Stereo,
@@ -130,21 +136,57 @@ public:
 	class Info
 	{
 	public:
-		Audio::Format format;
-		Audio::Encoding encoding;
+		Format format;
+		Encoding encoding;
 		unsigned rate;
 		unsigned order;
 		char *annotation;
 	};
 
-	static CCXX_EXPORT(bool) isMono(Encoding encoding);
-	static CCXX_EXPORT(bool) isStereo(Encoding encoding);
-	static CCXX_EXPORT(Rate) getRate(Encoding encoding);
-	static CCXX_EXPORT(int) getFrame(Encoding encoding, int samples = 0);
-	static CCXX_EXPORT(int) getCount(Encoding encoding);
-	static CCXX_EXPORT(unsigned long) toSamples(Encoding encoding, size_t bytes);
-	static CCXX_EXPORT(unsigned long) toBytes(Encoding encoding, unsigned long samples);
-	static CCXX_EXPORT(void) fill(unsigned char *addr, int samples, Encoding encoding);
+	static bool isMono(Encoding encoding);
+	static bool isStereo(Encoding encoding);
+	static Rate getRate(Encoding encoding);
+
+	/**
+	 * Returns the number of bytes in a sample frame for the given
+	 * encoding type, rounded up to the nearest integer.  A frame
+	 * is defined as the minimum number of bytes necessary to
+	 * create a point or points in the output waveform for all
+	 * output channels.  For example, 16-bit mono PCM has a frame
+	 * size of two (because those two bytes constitute a point in
+	 * the output waveform).  GSM has it's own definition of a
+	 * frame which involves decompressing a sequence of bytes to
+	 * determine the final points on the output waveform.  The
+	 * minimum number of bytes you can feed to the decompression
+	 * engine is 32.5 (260 bits), so this function will return 33
+	 * (because we round up) given an encoding type of GSM.  Other
+	 * compressed encodings will return similar results.  Be
+	 * prepared to deal with nonintuitive return values for
+	 * rare encodings.
+	 *
+	 * @param encoding The encoding type to get the frame size for.
+	 * @param samples Reserved.  Use zero.
+	 *
+	 * @return The number of bytes in a frame for the given encoding.
+	 */
+	static int getFrame(Encoding encoding, int samples = 0);
+
+	/**
+	 * Returns the number of samples in all channels for a frame
+	 * in the given encoding.  For example, pcm32Stereo has a
+	 * frame size of 8 bytes: Note that different codecs have
+	 * different definitions of a frame - for example, compressed
+	 * encodings have a rather large frame size relative to the
+	 * sample size due to the way bytes are fed to the
+	 * decompression engine.
+	 *
+	 * @param encoding The encoding to calculate the frame sample count for.
+	 * @return samples The number of samples in a frame of the given encoding.
+	 */
+	static int getCount(Encoding encoding);
+	static unsigned long toSamples(Encoding encoding, size_t bytes);
+	static unsigned long toBytes(Encoding encoding, unsigned long samples);
+	static void fill(unsigned char *addr, int samples, Encoding encoding);
 };
 
 /**
@@ -163,6 +205,8 @@ private:
 	Info info;
 	unsigned long header;		// offset to start of audio
 	unsigned long minimum;		// minimum sample size required
+	unsigned long length;           // current size of file,
+					// including header
 
 	void initialize(void);
 	void getWaveFormat(int size);
@@ -179,16 +223,63 @@ protected:
 	virtual bool afCreate(const char *path);
 	virtual bool afOpen(const char *path);
 	virtual bool afPeek(unsigned char *data, unsigned size);
+
+	/**
+	 * Read a given number of bytes from the file, starting from
+	 * the current file pointer.  May be overridden by derived
+	 * classes.
+	 *
+	 * @param data A pointer to the buffer to copy the bytes to.
+	 * @param size The number of bytes to read.
+	 * @return The number of bytes read, or -1 if an error occurs.
+	 * On UNIX platforms, use strerror(errno) to get the
+	 * human-readable error string or
+	 * FormatMessage(GetLastError()) on Windows platforms.
+	 */
 	virtual int afRead(unsigned char *data, unsigned size);
+
+	/** 
+	 * Write a number of bytes into the file at the current file
+	 * pointer.  May be overridden by derived classes.
+	 * 
+	 * @param data A pointer to the buffer with the bytes to write.
+	 * @param size The number of bytes to write from the buffer.
+	 * @return The number of bytes written, or -1 if an error
+	 * occurs.  On UNIX platforms, use strerror(errno) to get the
+	 * human-readable error string or
+	 * FormatMessage(GetLastError()) on Windows platforms.
+	 */
 	virtual int afWrite(unsigned char *data, unsigned size);
+
+	/**
+	 * Seek to the given position relative to the start of the
+	 * file and set the file pointer.  This does not use 64-bit
+	 * clean seek functions, so seeking to positions greater than
+	 * (2^32)-1 will result in undefined behavior.
+	 * 
+	 * @param pos The position to seek to.
+	 * @return true if successful, false otherwise.
+	 */
 	virtual bool afSeek(unsigned long pos);
 	virtual void afClose(void);
 
 	virtual char *getContinuation(void)
 		{return NULL;};
 
+	/**
+	 * Return a human-readable error message given a numeric error
+	 * code of type Audio::Error.
+	 *
+	 * @param err The numeric error code to translate.
+	 * @return A pointer to a character string containing the
+	 * human-readable error message.
+	 */
+	const char * getErrorStr(Error err);
 	Error setError(Error err);
 
+	unsigned long getHeader(void)
+		{return header;};
+
 	unsigned short getShort(unsigned char *data);
 	void setShort(unsigned char *data, unsigned short value);
 	unsigned long getLong(unsigned char *data);
@@ -201,22 +292,131 @@ public:
 	AudioFile()
 		{initialize();};
 
-	~AudioFile()
+	virtual ~AudioFile()
 		{clear();};
 
+	/**
+	 * Open an audio file and associate it with this object.
+	 * Called implicitly by the two-argument version of the
+	 * constructor.
+	 *
+	 * @param fname The name of the file to open.  Don't forget to
+	 * double your backslashes for DOS-style pathnames.
+	 */
 	void open(const char *fname);
+
+	/**
+	 * Create a new audio file and associate it with this object.
+	 * Called implicitly by the three-argument version of the
+	 * constructor.
+	 *
+	 * @param fname The name of the file to open.
+	 * @param info The type of the audio file to be created.
+	 */
 	void create(const char *fname, Info *info);
+
+	/**
+	 * Close an object associated with an open file.  This
+	 * updates the header metadata with the file length if the
+	 * file length has changed.
+	 */
 	void close(void);
+
+	/**
+	 * Clear the AudioFile structure.  Called by
+	 * AudioFile::close().  Sets all fields to zero and deletes
+	 * the dynamically allocated memory pointed to by the pathname
+	 * and info.annotation members.  See AudioFile::initialize()
+	 * for the dynamic allocation code.
+	 */
 	void clear(void);
+
+	/**
+	 * Retrieve bytes from the file into a memory buffer.  This
+	 * increments the file pointer so subsequent calls read further
+	 * bytes.  If you want to read a number of samples rather than
+	 * bytes, use getSamples().
+	 *
+	 * @param addr A pointer to the memory area to copy the samples to.
+	 * @param len The number of bytes (not samples) to copy.
+	 * @return The number of bytes (not samples) read.  Returns -1
+	 * if no bytes are read and an error occurs.
+	 */
 	int getBuffer(void *addr, unsigned len);
+
+	/**
+	 * Insert bytes into the file from a memory buffer.  This
+	 * increments the file pointer so subsequent calls append
+	 * further samples.  If you want to write a number of samples
+	 * rather than bytes, use putSamples().
+	 *
+	 * @param attr A pointer to the memory area to append the samples
+	 * from.
+	 * @param len The number of bytes (not samples) to append.
+	 * @return The number of bytes (not samples) read.  Returns -1
+	 * if an error occurs and no bytes are written.
+	 */
 	int putBuffer(void *attr, unsigned len);
+
+	/**
+	 * Retrieve samples from the file into a memory buffer.  This
+	 * increments the file pointer so subsequent calls read
+	 * further samples.  If a limit has been set using setLimit(),
+	 * the number of samples read will be truncated to the limit
+	 * position.  If you want to read a certain number of bytes
+	 * rather than a certain number of samples, use getBuffer().
+	 *
+	 * @param addr A pointer to the memory area to copy the samples to.
+	 * @param samples The number of samples to read.
+	 * @return errSuccess if successful, !errSuccess if
+	 * error.  Use getErrorStr() to retrieve the human-readable
+	 * error string.
+	 */
 	Error getSamples(void *addr, unsigned samples);
+
+	/**
+	 * Insert samples into the file from a memory buffer.  This
+	 * increments the file pointer so subsequent calls append
+	 * further samples.  If you want to write a certain number of
+	 * bytes rather than a certain number of samples, use
+	 * putBuffer().
+	 *
+	 * @param addr A pointer to the memory area to append the samples
+	 * from.
+	 * @param samples The number of samples (not bytes) to append.
+	 * @return errSuccess if successful, !errSuccess if
+	 * error.  Use getErrorStr() to retrieve the human-readable
+	 * error string.
+	 */
 	Error putSamples(void *addr, unsigned samples);
 	Error skip(long samples);
 	Error setPosition(unsigned long samples = ~0l);
 	Error setLimit(unsigned long samples = 0l);
 	Error getInfo(Info *info);
 	Error setMinimum(unsigned long samples);
+
+	/**
+	 * Get the current file pointer in bytes relative to the start
+	 * of the file.  See getPosition() to determine the position
+	 * relative to the start of the sample buffer.
+	 *
+	 * @return The current file pointer in bytes relative to the
+	 * start of the file.  Returns 0 if the file is not open, is
+	 * empty, or an error has occured.
+	 */
+	unsigned long getAbsolutePosition(void);
+
+	/**
+	 * Get the current file pointer in samples relative to the
+	 * start of the sample buffer.  Note that you must multiply
+	 * this result by the result of a call to
+	 * toBytes(info.encoding, 1) in order to determine the offset
+	 * in bytes.
+	 *
+	 * @return the current file pointer in samples relative to the
+	 * start of the sample buffer.  Returns 0 if the file is not
+	 * open, is empty, or an error has occured.
+	 */
 	unsigned long getPosition(void);
 	virtual bool isOpen(void);
 	virtual bool hasPositioning(void)
@@ -350,15 +550,16 @@ class CCXX_CLASS_EXPORT AudioCodec : public AudioSample
 {
 protected:
 	AudioCodec(unsigned frames, Encoding encoding, unsigned rate);
+	virtual ~AudioCodec() {};
 	unsigned length;
 
 public:
 	/**
 	 * Encode a linear sample frame into the codec sample buffer.
 	 *
-	 * @param linear sample buffer to encode.
-	 * @param number of samples, or use default count.
-	 * @param channel number or 0 for all channels.
+	 * @param buffer linear sample buffer to encode.
+	 * @param samples The number of samples, or use default count.
+	 * @param channel The channel number, or 0 for all channels.
 	 */
 	virtual void encode(unsigned short *buffer, unsigned samples = 0, int channel = 0) = 0;
 
@@ -367,7 +568,7 @@ public:
 	 * conversion is supported.
 	 *
 	 * @return true if successful.
-	 * @param object source.
+	 * @param sample object source.
 	 */
 	virtual bool encode(AudioSample *sample) = 0;
 
@@ -375,8 +576,8 @@ public:
 	 * Decode the sample frame into linear samples.
 	 *
 	 * @return number of samples extracted.
-	 * @param sample buffer to save linear samples into.
-	 * @param channel number or 0 for all channels.
+	 * @param buffer sample buffer to save linear samples into.
+	 * @param channel channel number or 0 for all channels.
 	 */
 	virtual unsigned decode(unsigned short *buffer, int channel = 0) = 0;
 
@@ -384,7 +585,7 @@ public:
 	 * Decode into an :AudioSample" if the conversion is supported.
 	 *
  	 * @return true if successful.
-	 * @param object to save into.
+	 * @param sample object to save into.
 	 */
 	virtual bool decode(AudioSample *sample) = 0;
 
@@ -394,8 +595,8 @@ public:
 	 * sizes.
 	 *
 	 * @return number of bytes actually read into codec workspace.
-	 * @param pointer to byte data stream.
-	 * @param number of data bytes.
+	 * @param data pointer to byte data stream.
+	 * @param length number of data bytes.
 	 */
 	virtual unsigned load(unsigned char *data, unsigned length);
 
@@ -405,8 +606,8 @@ public:
 	 * sizes.
 	 *
 	 * @return number of bytes actually written to workspace.
-	 * @param pointer to byte data stream.
-	 * @param size of byte data stream.
+	 * @param data pointer to byte data stream.
+	 * @param length size of byte data stream.
 	 */
 	virtual unsigned save(unsigned char *data, unsigned length);
 
@@ -463,7 +664,7 @@ public:
 	 * @return samples per second.
 	 */
 	virtual unsigned getRate(void)
-		{Audio::getRate(getEncoding());};
+		{return Audio::getRate(getEncoding());};
 
 	/**
 	 * Get a logical name for this codec.
@@ -477,7 +678,7 @@ public:
 	 * an object of "frame" size, as a multiple of the min. frame
 	 * size.  This is usually just a new xxxCodec(frames * minframe).
 	 *
-	 * @param number of "minframe" sample frames in transcoder.
+	 * @param frames number of "minframe" sample frames in transcoder.
 	 */
 	virtual AudioCodec *getCodec(unsigned frames) = 0;
 
@@ -486,10 +687,10 @@ public:
 	 * creates a tone buffer (AudioTone-like sample) encoded in
 	 * the format of the given codec.
 	 *
-	 * @param first tone frequency.
-	 * @param second tone frequency.
-	 * @param max frames of sample.
-         * @param max frames of tone.
+	 * @param f1 first tone frequency.
+	 * @param f2 second tone frequency.
+	 * @param maxsize max frames of sample.
+         * @param maxtone max frames of tone.
 	 * @return audio sample in encoded format.
 	 */
 	virtual AudioSample *getTone(unsigned f1, unsigned f2, unsigned maxsize, unsigned maxtone = 0) = 0;
@@ -514,15 +715,15 @@ public:
 	/**
 	 * Locate a codec registry object.
 	 *
-	 * @param encoding of codec registry to locate.
-	 * @param optional sample rate of codec registry to locate.
+	 * @param encoding encoding of codec registry to locate.
+	 * @param rate optional sample rate of codec registry to locate.
 	 */
 	static AudioRegisterCodec *findRegistration(Encoding encoding, unsigned rate = 0);
 
 	/**
 	 * Locate a codec registry object by name.
 	 *
-	 * @param name of codec registry to locate.
+	 * @param name name of codec registry to locate.
 	 */
 	static AudioRegisterCodec *findRegistration(const char *name);
 
@@ -530,9 +731,9 @@ public:
 	 * Create a new codec object from the registry to use for
 	 * transcoding.
 	 *
-	 * @param encoding of codec to create.
-	 * @param optional sample rate of codec being created.
-	 * @param sample frames (multiples of minframe) or 0 for defframe.
+	 * @param encoding encoding of codec to create.
+	 * @param rate optional sample rate of codec being created.
+	 * @param frames sample frames (multiples of minframe) or 0 for defframe.
 	 */
 	static AudioCodec *findCodec(Encoding encoding, unsigned rate = 0, unsigned frames = 0);
 
@@ -540,8 +741,8 @@ public:
 	 * Create a new codec object from a named registry to use for
 	 * transcoding.
 	 *
-	 * @param name of codec to create.
-	 * @param frame count to process or 0 for default frame size.
+	 * @param name name of codec to create.
+	 * @param frame frame count to process or 0 for default frame size.
 	 */
 	static AudioCodec *findCodec(const char *name, unsigned frame = 0);
 };
@@ -596,6 +797,7 @@ protected:
 
 public:
 	AudioCopy(unsigned frame, Encoding encoding = pcm16Mono, unsigned rate = 8000);
+	virtual ~AudioCopy() {};
 	bool copy(void);
 
 	inline bool isEmpty(void)
