@@ -40,6 +40,7 @@
 
 #include <cc++/config.h>
 #include <cc++/strchar.h>
+#include <cc++/export.h>
 #include "private.h"
 #include "audio.h"
 
@@ -47,6 +48,28 @@
 namespace ost {
 #endif
 
+static const char * const ErrorStrs[] = {
+	"errSuccess",
+	"errReadLast",
+	"errNotOpened",
+	"errEndOfFile",
+	"errStartOfFile",
+	"errRateInvalid",
+	"errEncodingInvalid",
+	"errReadInterrupt",
+	"errWriteInterrupt",
+	"errReadFailure",
+	"errWriteFailure",
+	"errReadIncomplete",
+	"errWriteIncomplete",
+	"errRequestInvalid",
+	"errTOCFailed",
+	"errStatFailed",
+	"errInvalidTrack",
+	"errPlaybackFailed",
+	"errNotPlaying"
+};
+
 void AudioFile::setShort(unsigned char *data, unsigned short val)
 {
 	if(info.order == __BIG_ENDIAN)
@@ -100,7 +123,7 @@ unsigned long AudioFile::getLong(unsigned char * data)
 
 AudioFile::AudioFile(const char *name, unsigned long sample)
 {
-	memset(&info, 0, sizeof(info));
+	memset(&this->info, 0, sizeof(this->info));
 	pathname = NULL;
 	initialize();
 	AudioFile::open(name);
@@ -112,7 +135,7 @@ AudioFile::AudioFile(const char *name, unsigned long sample)
 
 AudioFile::AudioFile(const char *name, Info *info, unsigned long samples)
 {
-	memset(&info, 0, sizeof(info));
+	memset(&this->info, 0, sizeof(this->info));
 	pathname = NULL;
 	initialize();
 	AudioFile::create(name, info);
@@ -124,13 +147,14 @@ AudioFile::AudioFile(const char *name, Info *info, unsigned long samples)
 
 void AudioFile::create(const char *name, Info *myinfo)
 {
+	int pcm = 0;
 	unsigned char aufile[24];
 	unsigned char riffhdr[40];
-	unsigned int count;
 
 	if(!afCreate(name))
 		return;
 
+	memset(riffhdr, 0, sizeof(riffhdr));
 	memcpy(&info, myinfo, sizeof(info));
 	info.annotation = NULL;
 	pathname = new char[strlen(name) + 1];
@@ -145,6 +169,18 @@ void AudioFile::create(const char *name, Info *myinfo)
 	{
 	case wave:
 	case riff:
+		/* 
+		 * RIFF header
+		 *
+		 * 12 bytes
+		 *
+		 * 0-3: RIFF magic "RIFF" (0x52 49 46 46)
+		 *
+		 * 4-7: RIFF header chunk size 
+		 *      (4 + (8 + subchunk 1 size) + (8 + subchunk 2 size))
+		 *
+		 * 8-11: WAVE RIFF type magic "WAVE" (0x57 41 56 45)
+		 */
 		if(!info.order)
 			info.order = __LITTLE_ENDIAN;
 		if(info.order == __LITTLE_ENDIAN)
@@ -154,7 +190,7 @@ void AudioFile::create(const char *name, Info *myinfo)
 		if(!info.rate)
 			info.rate = Audio::getRate(info.encoding);
 
-		header = 46;
+		header = 0;
 
 		memset(riffhdr + 4, 0xff, 4);
 		strncpy((char *)riffhdr + 8, "WAVE", 4);
@@ -164,65 +200,169 @@ void AudioFile::create(const char *name, Info *myinfo)
 			return;
 		}
 
+		/* 
+		 * Subchunk 1: WAVE metadata
+		 *
+		 * Length: 24+
+		 *
+		 * (offset from start of subchunk 1) startoffset-endoffset
+		 *
+		 * (0) 12-15: WAVE metadata magic "fmt " (0x 66 6d 74 20)
+		 *
+		 * (4) 16-19: subchunk 1 size minus 8.  0x10 for PCM.
+		 *
+		 * (8) 20-21: Audio format code.  0x01 for linear PCM.
+		 * More codes here:
+		 * http://www.goice.co.jp/member/mo/formats/wav.html
+		 *
+		 * (10) 22-23: Number of channels.  Mono = 0x01,
+		 * Stereo = 0x02.
+		 *
+		 * (12) 24-27: Sample rate in samples per second. (8000, 
+		 * 44100, etc)
+		 *
+		 * (16) 28-31: Bytes per second = SampleRate *
+		 * NumChannels * (BitsPerSample / 8)
+		 *
+		 * (20) 32-33: Block align (the number of bytes for
+		 * one multi-channel sample) = Numchannels *
+		 * (BitsPerSample / 8)
+		 *
+		 * (22) 34-35: Bits per single-channel sample.  8 bits
+		 * per channel sample = 0x8, 16 bits per channel
+		 * sample = 0x10
+		 *
+		 * (24) 36-37: Size of extra PCM parameters for
+		 * non-PCM formats.  If a PCM format code is
+		 * specified, this doesn't exist.
+		 *
+		 * Subchunk 3: Optional 'fact' subchunk for non-PCM formats
+		 * (26) 38-41: WAVE metadata magic 'fact' (0x 66 61 63 74)
+		 * 
+		 * (30) 42-45: Length of 'fact' subchunk, not
+		 * including this field and the fact
+		 * identification field (usually 4)
+		 *
+		 * (34) 46-49: ??? sndrec32.exe outputs 0x 00 00 00 00
+		 * here.  See 
+		 * http://www.health.uottawa.ca/biomech/csb/ARCHIVES/riff-for.txt
+		 *
+		 */
+
+		memset(riffhdr, 0, sizeof(riffhdr));
 		strncpy((char *)riffhdr, "fmt ", 4);
-		setLong(riffhdr + 4, 18);
-		setShort(riffhdr + 8, 1);
+		// FIXME A bunch of the following only works for PCM
+		// and mulaw/alaw, so you'll probably have to fix it
+		// if you want to use one of the other formats.
+		if(info.encoding < cdaStereo)
+			setLong(riffhdr + 4, 18);
+		else
+			setLong(riffhdr + 4, 16);
+
+		setShort(riffhdr + 8, 0x01); // default in case invalid encoding specified
 		if(isMono(info.encoding))
 			setShort(riffhdr + 10, 1);
 		else
 			setShort(riffhdr + 10, 2);
 		setLong(riffhdr + 12, info.rate);
 		setLong(riffhdr + 16, toBytes(info.encoding, info.rate));
-		setLong(riffhdr + 20, toBytes(info.encoding, info.rate) / 8);
-		setShort(riffhdr + 24, 0);
+		setShort(riffhdr + 20, toBytes(info.encoding, 1));
+		setShort(riffhdr + 22, 0);
 
 		switch(info.encoding)
 		{
 		case pcm8Mono:
 		case pcm8Stereo:
-			setShort(riffhdr + 24, 8);
+			setShort(riffhdr + 22, 8);
+			pcm = 1;
 			break;
 		case pcm16Mono:
 		case pcm16Stereo:
 		case cdaMono:
 		case cdaStereo:
-			setShort(riffhdr + 24, 16);
+			setShort(riffhdr + 22, 16);
+			pcm = 1;
 			break;
 		case pcm32Mono:
 		case pcm32Stereo:
-			setShort(riffhdr + 24, 32);
+			setShort(riffhdr + 22, 32);
+			pcm = 1;
 			break;
 		case alawAudio:
 			setShort(riffhdr + 8, 6);
+			setShort(riffhdr + 22, 8);
 			break;
 		case mulawAudio:
 			setShort(riffhdr + 8, 7);
+			setShort(riffhdr + 22, 8);
 			break;
+
+			// FIXME I'm pretty sure these are supposed to
+			// be writing to offset 22 instead of 24...
 		case okiADPCM:
 			setShort(riffhdr + 8, 0x10);
+			setShort(riffhdr + 24, 4);
 			break;
 		case voxADPCM:
 			setShort(riffhdr + 8, 0x17);
+			setShort(riffhdr + 24, 4);
 			break;
 		case g721ADPCM:
 			setShort(riffhdr + 8, 0x40);
+			setShort(riffhdr + 24, 4);
 			break;
 		case g722Audio:
 			setShort(riffhdr + 8, 0x64);
+			setShort(riffhdr + 24, 8);
 			break;
 		case gsmVoice:
 			setShort(riffhdr + 8, 0x31);
+			setShort(riffhdr + 24, 260);
 			break;
 		case g723_3bit:
+			setShort(riffhdr + 8, 0x14);
+			setShort(riffhdr + 24, 3);
+			break;
 		case g723_5bit:
 			setShort(riffhdr + 8, 0x14);
-		}
-		if(afWrite(riffhdr, 26) != 26)
-		{
-			AudioFile::close();
-			return;
+			setShort(riffhdr + 24, 5);
 		}
 
+		if(pcm == 0) {
+			setShort(riffhdr + 24, 0);
+			strncpy((char *)(riffhdr + 26), "fact", 4);
+			setLong(riffhdr + 30, 4);
+			setLong(riffhdr + 34, 0);
+			if(afWrite(riffhdr, 38) != 38)
+			{
+				AudioFile::close();
+				return;
+			}
+		}
+		else {
+			if(afWrite(riffhdr, 24) != 24)
+			{
+				AudioFile::close();
+				return;
+			}
+		}
+		
+		/* 
+		 * Subchunk 2: data subchunk
+		 *
+		 * Length: 8+
+		 *
+		 * (0) 36-39: data subchunk magic "data" (0x 64 61 74 61)
+		 *
+		 * (4) 40-43: subchunk 2 size = 
+		 * NumSamples * NumChannels * (BitsPerSample / 8)
+		 * Note that this does not include the size of this
+		 * field and the previous one.
+		 *
+		 * (8) 44+: Samples
+		 */
+
+		memset(riffhdr, 0, sizeof(riffhdr));
 		strncpy((char *)riffhdr, "data", 4);
 		memset(riffhdr + 4, 0xff, 4);
 		if(afWrite(riffhdr, 8) != 8)
@@ -230,6 +370,9 @@ void AudioFile::create(const char *name, Info *myinfo)
 			AudioFile::close();
 			return;
 		}
+		
+		header = getAbsolutePosition();
+		length = getAbsolutePosition();
 		break;
 
 	case sun:
@@ -295,13 +438,14 @@ void AudioFile::create(const char *name, Info *myinfo)
 		if(info.annotation)
 			afWrite((unsigned char *)info.annotation,
 				strlen(info.annotation) + 1);
+		header = getAbsolutePosition();
+		length = getAbsolutePosition();
 	}
 }
 
 void AudioFile::getWaveFormat(int request)
 {
 	unsigned char filehdr[24];
-	unsigned int count;
 	int bitsize;
 	int channels;
 
@@ -389,7 +533,6 @@ void AudioFile::open(const char *name)
 {
 	unsigned char filehdr[24];
 	unsigned int count;
-	unsigned bitsize;
 	char *ext;
 	unsigned channels;
 
@@ -419,6 +562,12 @@ void AudioFile::open(const char *name)
 		return;
 	}
 
+	if(!stricmp(ext, ".sw") || !stricmp(ext, ".raw"))
+	{
+		info.encoding = pcm16Mono;
+		return;
+	}
+
 	if(!stricmp(ext, ".vox"))
 	{
 		info.encoding = voxADPCM;
@@ -564,8 +713,38 @@ void AudioFile::open(const char *name)
 
 void AudioFile::close(void)
 {
-	unsigned long pos;
+	unsigned char buf[58];
+	if(! isOpen())
+		return;
+	if(! afSeek(0)) {
+		return;
+	}
 
+	if(-1 == afRead(buf, 58)) {
+		return;
+	}
+	afSeek(0);
+
+	switch(info.format) {
+	case riff:
+		// RIFF header
+		setLong(buf + 4, length - 8);
+
+		// If it's a non-PCM datatype, the offsets are a bit
+		// different for subchunk 2.
+		if(info.encoding < cdaStereo)
+			setLong(buf + 54, length - header);
+		else
+			setLong(buf + 40, length - header); 
+
+		if(afWrite(buf, 58) != 58)
+			return;
+
+		break;
+	case sun:
+		// FIXME not implemented
+		break;
+	}
 	afClose();
 	clear();
 }
@@ -605,6 +784,11 @@ Audio::Error AudioFile::setError(Error err)
 	return err;
 }
 
+const char * AudioFile::getErrorStr(Error err)
+{
+	return ErrorStrs[err];
+}
+
 Audio::Error AudioFile::setMinimum(unsigned long samples)
 {
 	if(!isOpen())
@@ -635,7 +819,7 @@ int AudioFile::getBuffer(void *addr, unsigned bytes)
 			return xfer;
 		}
 		xfer += count;
-		if(count == bytes)
+		if(count == (int)bytes)
 			return xfer;
 		fname = getContinuation();
 		if(!fname)
@@ -653,7 +837,7 @@ Audio::Error AudioFile::getSamples(void *addr, unsigned request)
 {
 	char *fname;
 	unsigned char *caddr = (unsigned char *)addr;
-	int count, bytes;
+	int count;
 	for(;;)
 	{
 		int bytes = toBytes(info.encoding, request);
@@ -668,7 +852,7 @@ Audio::Error AudioFile::getSamples(void *addr, unsigned request)
 
 		if(count > 0)
 		{
-			caddr += toBytes(info.encoding, count);
+			caddr += count;
 			count = request - toSamples(info.encoding, count);
 		}
 		else
@@ -692,7 +876,19 @@ Audio::Error AudioFile::getSamples(void *addr, unsigned request)
 
 int AudioFile::putBuffer(void *addr, unsigned len)
 {
-	return afWrite((unsigned char *)addr, len);
+	int count;
+
+	count = afWrite((unsigned char *)addr, len);
+	if(count == (int)len) {
+		length += count;
+		return count;
+	}
+	if(count < 1)
+		return count;
+	else {
+ 		length += count;
+		return count;
+	}
 }
 
 Audio::Error AudioFile::putSamples(void *addr, unsigned samples)
@@ -703,22 +899,29 @@ Audio::Error AudioFile::putSamples(void *addr, unsigned samples)
 		return setError(errRequestInvalid);
 
 	count = afWrite((unsigned char *)addr, bytes);
-	if(count == bytes)
+	if(count == bytes) {
+		length += count;
 		return errSuccess;
-
+	}
 	if(count < 1)
 		return errWriteFailure;
-	else
+	else {
+		length += count;
 		return errWriteIncomplete;
+	}
 }
 
 Audio::Error AudioFile::skip(long samples)
 {
-	long pos = getPosition() + samples;
+	unsigned long orig = getPosition();
+	unsigned long pos = orig + samples;
 	if(pos < 0)
 		pos = 0;
 
 	setPosition(pos);
+	if(orig < getPosition())
+		length += (getPosition() - orig);
+
 	return errSuccess;
 }
 
